from __future__ import print_function

import os
import sys

from mrbob.bobexceptions import ValidationError
from mrbob.configurator import Configurator, Question

from canari.commands.common import canari_main
from canari.question import parse_bool
from canari.commands.framework import SubCommand, Argument

__author__ = 'Nadeem Douba'
__copyright__ = 'Copyright 2012, Canari Project'
__credits__ = []

__license__ = 'GPLv3'
__version__ = '0.2'
__maintainer__ = 'Nadeem Douba'
__email__ = 'ndouba@gmail.com'
__status__ = 'Development'


def check_port(configurator, question, answer):
    try:
        if 1 < int(answer) <= 65535:
            return answer
        raise ValueError("Port number needs to be between 1 and 65535.")
    except ValueError as e:
        raise ValidationError(e)


def check_gid(configurator, question, answer):
    try:
        import grp
        return grp.getgrnam(answer).gr_gid
    except KeyError as e:
        raise ValidationError(e)


def check_uid(configurator, question, answer):
    try:
        import pwd
        return pwd.getpwnam(answer).pw_uid
    except KeyError as e:
        raise ValidationError(e)


def validate_path(configurator, question, answer):
    if not os.path.lexists(answer):
        raise ValidationError("Could not find file %r" % answer)
    return os.path.realpath(answer)


def check_mkdir(configurator, question, answer):
    try:
        from stat import ST_MODE, S_ISDIR
        if not os.path.lexists(answer):
            os.makedirs(answer)
        elif not S_ISDIR(os.stat(answer)[ST_MODE]):
            raise ValidationError("Invalid path (%s): path must point to a directory not a file." % answer)
        return os.path.realpath(answer)
    except OSError as e:
        raise ValidationError(e)


def check_init_script(configurator, question, answer):
    try:
        from stat import ST_MODE, S_ISDIR
        if not os.path.lexists(answer):
            os.makedirs(answer)
        elif not S_ISDIR(os.stat(answer)[ST_MODE]):
            raise ValidationError("Invalid path (%s): path must point to a directory not a file." % answer)

        with open(os.path.join(answer, 'plume'), 'w'):
            answer = os.path.realpath(answer)
            configurator.target_directory = answer
            return answer
    except OSError as e:
        raise ValidationError(e)
    except IOError as e:
        raise ValidationError(e)


def configure_ssl(configurator, question, answer):
    answer = answer.lower()
    if answer not in ('0', '1', 'y', 'n'):
        raise ValidationError("Value must be either y/n.")
    answer = int(answer in ('1', 'y'))
    if answer:
        configurator.questions.extend([
            Question(
                    'plume.certificate',
                    'Please specify the path to the public key',
                    default='/etc/ssl/private/server.pem',
                    required=True,
                    post_ask_question='canari.commands.install_plume:validate_path'
            ),
            Question(
                    'plume.private_key',
                    'Please specify the path to the private key',
                    default='/etc/ssl/private/server.pem',
                    required=True,
                    post_ask_question='canari.commands.install_plume:validate_path'
            ),
        ])
    else:
        configurator.variables['plume.certificate'] = ''
        configurator.variables['plume.private_key'] = ''
    return answer


@SubCommand(
    canari_main,
    help='Sets up Canari Plume directory structure and configuration files.',
    description='Sets up Canari Plume directory structure and configuration files.'
)
@Argument(
    '--accept-defaults',
    '-y',
    help='Install Plume with all the defaults in non-interactive mode.',
    default=False,
    action='store_true'
)
def install_plume(opts):

    if not opts.accept_defaults:
        return install_wizard(opts)
    install_defaults(opts)


def install_defaults(opts):
    configurator = Configurator('canari.resources.templates:install_plume', '.',
                                {'non_interactive': True, 'remember_answers': False})

    configurator.variables['plume.venv'] = os.environ.get('VIRTUAL_ENV')
    if configurator.variables['plume.venv']:
        print(
            'Will use the virtual environment in %r to run Plume...' % configurator.variables['plume.venv'],
            file=sys.stderr
        )
    configurator.variables['plume.enable_ssl'] = 'n'
    print('Installing init script to /etc/init.d...', file=sys.stderr)
    configurator.variables['plume.init'] = check_init_script(configurator, '', '/etc/init.d')
    print('Creating Plume root directory at /var/plume...', file=sys.stderr)
    configurator.variables['plume.dir'] = check_mkdir(configurator, '', '/var/plume')
    print('The PID file will be at /var/run/plume.pid...', file=sys.stderr)
    configurator.variables['plume.run_dir'] = '/var/run'
    print('The log files will be at /var/log/plume.log...', file=sys.stderr)
    configurator.variables['plume.log_dir'] = '/var/log'
    configurator.variables['plume.user'] = check_uid(configurator, '', 'nobody')
    configurator.variables['plume.group'] = check_gid(configurator, '', 'nobody')
    print('The Plume server will under UID/GID=%s/%s...' % (
        configurator.variables['plume.user'], configurator.variables['plume.group']), file=sys.stderr)
    print('TLS will be disabled by default...', file=sys.stderr)
    configurator.variables['plume.certificate'] = ''
    configurator.variables['plume.private_key'] = ''

    configurator.ask_questions()
    configurator.render()
    finish(configurator)


def install_wizard(opts):
    configurator = Configurator('canari.resources.templates:install_plume', '.',
                                {'non_interactive': False, 'remember_answers': False})
    configurator.ask_questions()

    if os.environ.get('VIRTUAL_ENV'):
        run_venv = parse_bool(
            "--> Canari has detected that you're running this install script from within a virtualenv.\n"
            "--> Would you like to run Plume from this virtualenv (%r) as well?" % os.environ['VIRTUAL_ENV'], True)
        configurator.variables['plume.venv'] = os.environ['VIRTUAL_ENV'] if run_venv else False
    else:
        configurator.variables['plume.venv'] = None

    configurator.render()
    finish(configurator)


def finish(configurator):
    print('Writing canari.conf to %r...' % configurator.variables['plume.dir'], file=sys.stderr)

    # move the canari.conf file from the init.d directory to the plume content directory
    src_file = os.path.join(configurator.variables['plume.init'], 'canari.conf')
    dst_file = os.path.join(configurator.variables['plume.dir'], 'canari.conf')

    if src_file != dst_file:
        with open(src_file) as src:
            with open(dst_file, 'w') as dst:
                dst.write(src.read())
        os.unlink(src_file)

    os.chmod(os.path.join(configurator.variables['plume.init'], 'plume'), 0o755)

    print('done!', file=sys.stderr)



#!/usr/bin/env python

import os
import sys


from setuptools import setup, find_packages


long_description = """
Canari Framework 3
==================

|Build Status| |Doc Status|

Welcome to the Canari 3 repository - the next generation Maltego rapid
transform development framework which allows you to rapidly prototype,
package, and distribute Maltego local and remote transforms. Please
visit the `documentation <http://canari3.readthedocs.io/en/latest/>`__
site for a quick how-to and more in-depth information on the framework
itself.

Sneak Peek
----------

The following is an example of how easy it is to write a quick Maltego
transform in Canari 3:

.. code:: python


    from canari.maltego.entities import Phrase, Person

    class HelloWorld(Transform):
        \"\"\"This transform says hello to a person entity.\"\"\"

        # The transform input entity type.
        input_type = Person

        def do_transform(self, request, response, config):
            return response + Phrase("Hello " + request.entity.value)

Canari Docker
-------------

You can now dockerize your remote transform packages using
``canari dockerize-package``. This will create a Docker container that
runs Canari Plume fully configured with your remote transforms. You can
easily distribute this container to your Docker swarm. Check out the
documentation on Docker `website <http://docker.com>`__ for more
information on how containers work.

Bug Reports & Questions
-----------------------

Please use the issues page to log any bugs or questions regarding the
Canari Framework.

Kudos
-----

Kudos to our user community for making this release happen. A special
thanks to those of you who have supported the development of Canari 3 by
donating money at our crowd-funding pages. If you like this project,
please consider donating money to help accelerate development.

.. |Build Status| image:: https://circleci.com/gh/redcanari/canari3.svg?style=shield&circle-token=da787a222c75b0a739152d0aa92a9465f702bae6
.. |Doc Status| image:: https://readthedocs.org/projects/canari3/badge/?version=latest


"""

if os.path.exists('README.rst'):
    long_description = open('README.rst').read()

scripts = [
    'canari',
    'dispatcher'
]

extras = [
    'readline'
]

extras_require = [
    'mr.bob',
    'argparse',
    'flask',
    'Twisted',
    'pyopenssl',
    'service_identity',
    'pyasn1',
    'boto3'
]

requires = [
    'safedexml',
    'lxml',
    'six',
    'future'
]

if sys.platform == 'win32':
    requires.append('pyreadline')
    scripts.extend(['%s.bat' % s for s in scripts])
else:
    scripts.append('pysudo')

setup(
    name='canari',
    author='Nadeem Douba',
    version='3.3.0',
    author_email='ndouba@gmail.com',
    description='Canari Framework - Maltego rapid transform development and execution framework.',
    long_description=long_description,
    license='GPLv3',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    extras_require={
        'all': extras_require
    },
    install_requires=requires,
    dependency_links=[],
    url='https://github.com/redcanari/canari',
    scripts=[os.path.join('scripts', s) for s in scripts],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Other Audience',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Telecommunications Industry',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Server'
    ]
)



from __future__ import print_function

import os
import re
import sys
import tempfile
import time
from distutils.version import LooseVersion
from importlib import import_module
from zipfile import ZipFile, ZIP_STORED

from past.builtins import basestring
from safedexml import Model

from canari.maltego.configuration import (TransformSettings, CmdLineTransformPropertySetting, InputConstraint, Set,
                                          CmdParmTransformPropertySetting, CmdCwdTransformPropertySetting,
                                          CmdDbgTransformPropertySetting, MaltegoTransform,
                                          CmdLineTransformProperty, CmdCwdTransformProperty, CmdDbgTransformProperty,
                                          CmdParmTransformProperty, TransformSet, Transform, MaltegoServer, Protocol,
                                          Authentication, EntityCategory, Properties, attributes, fileobject, attr,
                                          MaltegoEntity)
from canari.question import parse_int
from canari.utils.common import find_dispatcher

__author__ = 'Nadeem Douba'
__copyright__ = 'Copyright 2012, Canari Project'
__credits__ = ['Nadeem Douba']

__license__ = 'GPLv3'
__version__ = '0.1'
__maintainer__ = 'Nadeem Douba'
__email__ = 'ndouba@gmail.com'
__status__ = 'Development'

__all__ = [
    'MaltegoDistribution',
    'MtzDistribution'
]


class MaltegoVersion:
    Radium = '3.2.0'
    Tungsten = '3.4.0'
    Carbon = '3.5.0'
    Chlorine = '3.6.0'


class MaltegoDistribution(object):

    def __init__(self, maltego_prefix=None, **kwargs):
        self._maltego_prefix = maltego_prefix if maltego_prefix else self._detect_settings_dir()
        self._maltego_config_dir = self.path_join(self.maltego_prefix, 'config', 'Maltego')
        self._seeds_dir = self.path_join(self.maltego_config_dir, 'Seeds')
        self._servers_dir = self.path_join(self.maltego_config_dir, 'Servers')
        self._entities_dir = self.path_join(self.maltego_config_dir, 'Entities')
        self._machines_dir = self.path_join(self.maltego_config_dir, 'Machines')
        self._icons_dir = self.path_join(self.maltego_config_dir, 'Icons')
        self._viewlets_dir = self.path_join(self.maltego_config_dir, 'Viewlets')
        self._transform_sets_dir = self.path_join(self.maltego_config_dir, 'TransformSets')
        self._transform_repositories_dir = self.path_join(self.maltego_config_dir, 'TransformRepositories')
        version = os.path.basename(self.maltego_prefix)
        self._version = LooseVersion(version[1:]) if version.startswith('v') else LooseVersion(version)
        self._machine_nbattr = self._get_machine_nbattr()

    @property
    def maltego_prefix(self):
        return self._maltego_prefix

    def _get_dir(self, dir_name, subdir_name=None, create=True):
        path = dir_name
        if subdir_name:
            path = os.path.join(dir_name, subdir_name)
        if not os.path.lexists(path):
            if create:
                os.makedirs(path, 0o755)
            else:
                raise OSError('Path does not exist: %r' % path)
        return path

    @property
    def viewlets_dir(self):
        return self._viewlets_dir

    @property
    def icons_dir(self):
        return self._icons_dir

    @property
    def icon_categories(self):
        return [p for p in os.listdir(self.icons_dir) if os.path.isdir(os.path.join(self.icons_dir, p))]

    @property
    def icon_files(self):
        return [os.path.join(p, i) for p, d, f in os.walk(self.icons_dir) for i in f if i]

    @property
    def icons(self):
        return self.icon_files

    @property
    def servers_dir(self):
        return self._servers_dir

    @property
    def server_files(self):
        return [os.path.join(self.servers_dir, s) for s in os.listdir(self.servers_dir) if s.endswith('.tas')]

    @property
    def servers(self):
        return self.server_files

    @property
    def maltego_config_dir(self):
        return self._maltego_config_dir

    @property
    def seeds_dir(self):
        return self._seeds_dir

    @property
    def entities_dir(self):
        return self._entities_dir

    @property
    def machine_files(self):
        return [os.path.join(self.machines_dir, m) for m in os.listdir(self.machines_dir) if m.endswith('.machine')]

    @property
    def machines(self):
        return self.machine_files

    @property
    def entity_files(self):
        return [os.path.join(p, i) for p, d, f in os.walk(self.entities_dir) for i in f if i.endswith('.entity')]

    @property
    def entities(self):
        return self.entity_files

    @property
    def transform_files(self):
        return [(os.path.join(p, t), '%ssettings' % os.path.join(p, t)) for p, d, f in
                os.walk(self.transform_repositories_dir) for t in f if t.endswith('.transform')]

    @property
    def transforms(self):
        return self.transform_files

    @property
    def transform_uuids(self):
        return [t.replace('.transform', '') for p, d, f in
                os.walk(self.transform_repositories_dir) for t in f if t.endswith('.transform')]

    @property
    def entity_categories(self):
        return [p for p in os.listdir(self.entities_dir)
                if os.path.isdir(os.path.join(self.entities_dir, p))]

    @property
    def transform_sets(self):
        return [p for p in os.listdir(self.transform_sets_dir)
                if os.path.isdir(os.path.join(self.transform_sets_dir, p))]

    @property
    def transform_repositories_dir(self):
        return self._transform_repositories_dir

    @property
    def transform_repositories(self):
        return [p for p in os.listdir(self.transform_repositories_dir)
                if os.path.isdir(os.path.join(self.transform_repositories_dir, p))]

    @property
    def transform_sets_dir(self):
        return self._transform_sets_dir

    @property
    def machines_dir(self):
        return self._machines_dir

    @property
    def seed_files(self):
        return [os.path.join(self.seeds_dir, s) for s in os.listdir(self.seeds_dir) if s.endswith('.seed')]

    @property
    def seeds(self):
        return self.seed_files

    @property
    def version(self):
        return self._version

    def get_transform_repository_dir(self, transform_repository_name):
        return self._get_dir(self.transform_repositories_dir, transform_repository_name,
                             False if self.version >= MaltegoVersion.Tungsten else True)

    def get_icon_category_dir(self, icon_category_name):
        return self._get_dir(self.icons_dir, icon_category_name,
                             False if self.version >= MaltegoVersion.Tungsten else True)

    def get_entity_category_dir(self, entity_category_name):
        return self._get_dir(self.entities_dir, entity_category_name,
                             False if self.version >= MaltegoVersion.Tungsten else True)

    def get_transform_set_dir(self, transform_set_name):
        return self._get_dir(self.transform_sets_dir, transform_set_name,
                             False if self.version >= MaltegoVersion.Tungsten else True)

    def add_server(self, server_name, **kwargs):
        pass

    def add_transform_to_server(self, server_name, transform_id):
        pass

    def remove_transform_from_server(self, server_name, transform_id):
        pass

    def add_transform_set(self, transform_set_name):
        if self.version >= MaltegoVersion.Tungsten:
            raise NotImplementedError('This version of Maltego uses encrypted configuration files and therefore does '
                                      'not support direct management of transform sets.')
        self.get_transform_set_dir(transform_set_name)

    def add_transform_repository(self, transform_repository_name):
        if self.version >= MaltegoVersion.Tungsten:
            raise NotImplementedError('This version of Maltego uses encrypted configuration files and therefore does '
                                      'not support direct management of transform repositories.')
        self.get_transform_repository_dir(transform_repository_name)

    def add_entity_category(self, entity_category_name):
        if self.version >= MaltegoVersion.Tungsten:
            raise NotImplementedError('This version of Maltego uses encrypted configuration files and therefore does '
                                      'not support direct management of entity categories.')
        self.get_entity_category_dir(entity_category_name)

    def add_icon_category(self, icon_category_name):
        if self.version >= MaltegoVersion.Tungsten:
            raise NotImplementedError('This version of Maltego uses encrypted configuration files and therefore does '
                                      'not support direct management of icon categories.')
        self.get_entity_category_dir(icon_category_name)

    def get_transforms_by_repository(self, transform_repository_name):
        return [t for t in self.transforms
                if t[0].startswith(self.get_transform_repository_dir(transform_repository_name))]

    def get_icons_by_category(self, icon_category_name):
        return [i for i in self.icon_files if i.startswith(self.get_icon_category_dir(icon_category_name))]

    def _iter_settings_dir(self, maltego_base_dir):
        vs = [i for i in os.listdir(maltego_base_dir) if os.path.isdir(os.path.join(maltego_base_dir, i)) and
                                                         os.path.isdir(os.path.join(maltego_base_dir, i, 'config'))]
        if len(vs) == 1:
            return os.path.join(maltego_base_dir, vs[0])
        elif vs:
            print('Multiple versions of Maltego detected: ', file=sys.stderr)
            r = parse_int('Please select which version you wish to use', ['Maltego %s' % i for i in vs])
            return os.path.join(maltego_base_dir, vs[int(r)])
        print('Could not automatically find Maltego\'s settings directory. '
              'Use the -w parameter to specify its location, instead.', file=sys.stderr)

    def _detect_settings_dir(self):
        if sys.platform.startswith('linux'):
            return self._iter_settings_dir(os.path.join(os.path.expanduser('~'), '.maltego'))
        elif sys.platform == 'darwin':
            return self._iter_settings_dir(
                os.path.join(os.path.expanduser('~'), 'Library', 'Application Support', 'maltego')
            )
        elif sys.platform == 'win32':
            return self._iter_settings_dir(os.path.join(os.environ['APPDATA'], '.maltego'))
        raise NotImplementedError('Unknown or unsupported OS: %r' % sys.platform)

    def add_transform_to_set(self, transform_id, transform_set):
        self.add_transform_set(transform_set)
        self.write_file(self.path_join(self.get_transform_set_dir(transform_set), transform_id), '')

    def remove_transform_from_set(self, transform_id, transform_set):
        transform_set_dir = self.get_transform_set_dir(transform_set)
        path = os.path.join(transform_set_dir, transform_id)
        self.remove_file(path)
        if not os.listdir(transform_set_dir):
            os.rmdir(transform_set_dir)

    def _get_module_author(self, module_name):
        return getattr(import_module(module_name), '__author__', '')

    def add_transform(self, working_dir, transform_repository, transform_class, server=None):
        transform_repository_dir = self.get_transform_repository_dir(transform_repository)

        transform = transform_class()
        name = '.'.join([transform_class.__module__, transform_class.__name__])
        author = transform.author or self._get_module_author(transform.__module__)
        transform_id = transform.name
        input_set = transform.transform_set
        input_entity = transform.input_type

        if transform.name in self.transform_uuids:
            print('WARNING: Previous declaration of %s in transform %s. Overwriting...' % (transform_id, name),
                  file=sys.stderr)
        else:
            print('Installing transform %s from %s...' % (transform_id, name), file=sys.stderr)

        sets = None
        if input_set:
            sets = Set(name=input_set)
            self.add_transform_to_set(transform_id, input_set)

        if server:
            self.add_transform_to_server(server, transform_id)

        transform_def = MaltegoTransform(
            name=transform_id,
            displayname=transform.display_name,
            author=author,
            description=transform.description,
            properties=(
                Properties() +
                CmdLineTransformProperty() +
                CmdCwdTransformProperty() +
                CmdDbgTransformProperty() +
                CmdParmTransformProperty()
            ),
            input=[InputConstraint(type=input_entity._type_)],
            sets=[sets]
        )

        self.write_file(
            self.path_join(transform_repository_dir, '%s.transform' % transform_id),
            transform_def
        )

        if not transform_def.sets:
            print('WARNING: Transform does not appear to be part of any Transform Sets (Perhaps an error?).',
                  file=sys.stderr)

        transform_settings_def = TransformSettings(properties=[
            CmdLineTransformPropertySetting(find_dispatcher()),
            CmdParmTransformPropertySetting(name),
            CmdCwdTransformPropertySetting(working_dir),
            CmdDbgTransformPropertySetting(transform.debug)
        ])

        self.write_file(
            self.path_join(transform_repository_dir, '%s.transformsettings' % transform_id),
            transform_settings_def
        )

    def remove_transform(self, transform_repository, transform, server=None):
        spec = transform.dotransform
        transform_repository_dir = self.get_transform_repository_dir(transform_repository)

        for transform_id, (input_set, input_entity) in zip(spec.uuids, spec.inputs):
            print('Removing %s from %s...' % (transform_id, transform_repository_dir), file=sys.stderr)
            self.remove_file(self.path_join(transform_repository_dir, '%s.transform' % transform_id))
            self.remove_file(self.path_join(transform_repository_dir, '%s.transformsettings' % transform_id))
            self.remove_transform_from_set(transform_id, input_set)
            if server:
                self.remove_transform_from_server(server, transform_id)

    def _get_machine_nbattr(self):
        nbattr = attributes()
        if self.version >= MaltegoVersion.Tungsten:
            return nbattr
        f = self.path_join(self.machines_dir, '.nbattrs')
        if os.path.exists(f):
            data = open(f).read()
            if data:
                nbattr = nbattr.parse(data)
        return nbattr

    def _parse_machine_script(self, contents):
        machine_sig = re.search('machine\s*\n*\((.+)\)\s*\n*{\s*\n*start\s*\n*{', contents, re.MULTILINE | re.DOTALL)
        if not machine_sig:
            return
        props = dict(re.findall(r'(?:(\w+)\s*:\s*)?["\']([^"\']+)["\']', machine_sig.groups()[0]))
        props['name'] = props.pop('')
        return props

    def remove_machine(self, filename):
        if self.version >= MaltegoVersion.Tungsten:
            raise NotImplementedError('This version of Maltego uses encrypted files and therefore does '
                                      'not support direct management of machine files.')
        filename = os.path.basename(filename)
        name = filename.replace('.machine', '')

        print('Uninstalling %s from %s...' % (filename, self.machines_dir), file=sys.stderr)
        if name in self._machine_nbattr.fileobjects:
            del self._machine_nbattr.fileobjects[name]

        self.remove_file(os.path.join(self.machines_dir, filename))

    def add_machine(self, filename, contents):
        if self.version >= MaltegoVersion.Tungsten:
            raise NotImplementedError('This version of Maltego uses encrypted files and therefore does '
                                      'not support direct management of machine files.')
        props = self._parse_machine_script(contents)
        if not props:
            return
        if props['name'] not in self._machine_nbattr.fileobjects:
            f = fileobject(name=props['name'])
            self._machine_nbattr += f
            for n, v in props.items():
                f += attr(name=n, stringvalue=v)
            f += attr(name='enabled', boolvalue=True)
            f += attr(name='readonly', boolvalue=False)

        dst = self.path_join(self.machines_dir, os.path.basename(filename))
        print('Installing machine %s to %s...' % (filename, dst), file=sys.stderr)
        self.write_file(dst, contents)

    def add_entity(self, entity):
        if isinstance(entity, basestring):
            entity = MaltegoEntity.parse(entity)
        if not isinstance(entity, MaltegoEntity):
            raise TypeError('Expected str or MaltegoEntity, not %s', type(entity).__name__)
        entity_category_dir = self.get_entity_category_dir(entity.category)
        entity_filename = '%s.entity' % self.path_join(entity_category_dir, entity.id)
        print('Installing entity %s to %s...' % (entity.id, entity_filename), file=sys.stderr)
        self.write_file(entity_filename, entity.render(fragment=True, pretty=True))

    def path_join(self, *args):
        return os.path.join(*args)

    def write_file(self, filename, contents):
        if self.version >= MaltegoVersion.Tungsten:
            raise NotImplementedError('This version of Maltego uses encrypted files and therefore does '
                                      'not support direct management of any files within the configuration directory.')
        with open(filename, 'w') as f:
            if isinstance(contents, Model):
                f.write(contents.render(fragment=True, pretty=True))
            else:
                f.write(contents)

    def read_file(self, filename):
        return open(filename).read()

    def _write_pending(self):
        nbattrs = self.path_join(self.machines_dir, '.nbattrs')
        with open(nbattrs, mode='w') as f:
            f.write(self._machine_nbattr.render(fragment=True, pretty=True))

    def remove_file(self, filename):
        if os.path.lexists(filename):
            os.unlink(filename)

    def __del__(self):
        if not hasattr(self, '_version') or self.version >= MaltegoVersion.Tungsten:
            return
        self._write_pending()


class MtzZipFile(object):

    def __init__(self, file_, mode="r", compression=ZIP_STORED, allowZip64=False, pwd=None):
        self._zipfile = None
        self.mode = mode
        self.filename = file_
        self._is_closed = False
        self.compression = ZIP_STORED
        self.allowZip64 = allowZip64
        if mode == "r":
            self._zipfile = ZipFile(file_, mode, compression, allowZip64)
        else:
            self._root_dir = tempfile.mkdtemp() + os.path.sep
            if self.mode == 'a':
                ZipFile(file_, 'r').extractall(self._root_dir, pwd)

    def namelist(self):
        if self._is_closed:
            raise RuntimeError('Attempt to read ZIP archive that was already closed')
        if self.mode == 'r':
            return self._zipfile.namelist()
        namelist = []
        for p, d, f in os.walk(self._root_dir):
            if f:
                for f_ in f:
                    f_ = os.path.join(p, f_)
                    namelist.append(self._get_archive_file(f_))
        return namelist

    def writestr(self, zinfo_or_arcname, bytes_):
        if self._is_closed:
            raise RuntimeError('Attempt to read ZIP archive that was already closed')
        if self.mode == 'r':
            return self._zipfile.writestr(zinfo_or_arcname, bytes_)
        realpath = self._get_real_file(zinfo_or_arcname)
        realdir = os.path.dirname(realpath)
        if not os.path.isdir(realdir):
            os.makedirs(realdir)
        with open(realpath, mode='w') as f:
            f.write(bytes_)

    def removefile(self, zinfo_or_arcname):
        if self._is_closed:
            raise RuntimeError('Attempt to remove file from ZIP archive that was already closed')
        if self.mode == 'r':
            return self._zipfile.writestr(zinfo_or_arcname, bytes)
        realpath = self._get_real_file(zinfo_or_arcname)
        if os.path.lexists(realpath):
            os.unlink(realpath)

    def read(self, name, pwd=None):
        if self._is_closed:
            raise RuntimeError('Attempt to read ZIP archive that was already closed')
        if self.mode == 'r':
            return self._zipfile.read(name, pwd)
        with open(self._get_real_file(name)) as f:
            return f.read()

    def _get_real_file(self, name):
        if sys.platform == 'win32':
            name = name.replace('/', os.path.sep)
        return os.path.join(self._root_dir, name)

    def _get_archive_file(self, name):
        name = name.replace(self._root_dir, '', 1)
        if sys.platform == 'win32':
            name = name.replace(os.path.sep, '/')
        return name

    def close(self):
        if self._is_closed or self.mode == 'r':
            return
        self._is_closed = True
        with ZipFile(self.filename, mode='w', compression=self.compression, allowZip64=self.allowZip64) as z:
            for p, d, f in os.walk(self._root_dir):
                if f:
                    for f_ in f:
                        name = os.path.join(p, f_)
                        z.write(name, self._get_archive_file(name))
                        os.unlink(name)

    def __del__(self):
        if not self._is_closed:
            self.close()


class MtzDistribution(MtzZipFile, MaltegoDistribution):

    def __init__(self, file_, mode="r", compression=ZIP_STORED, allowZip64=False):
        super(MtzDistribution, self).__init__(file_, mode, compression, allowZip64)
        self._maltego_config_dir = self.filename
        self._maltego_prefix = ''
        self._icons_dir = 'Icons'
        self._seeds_dir = 'Seeds'
        self._servers_dir = 'Servers'
        self._viewlets_dir = 'Viewlets'
        self._entities_dir = 'Entities'
        self._machines_dir = 'Machines'
        self._transform_sets_dir = 'TransformSets'
        self._transform_repositories_dir = 'TransformRepositories'
        self._version = NotImplemented
        self._transform_sets = {}
        self._servers = {}
        self._init_transform_sets()
        self._init_servers()

    def _init_servers(self):
        for s in self.namelist():
            if s.endswith('.tas'):
                self._append_server(s)

    def _append_server(self, s):
        s = MaltegoServer.parse(self.read(s))
        self.add_server(
            s.name,
            enabled=s.enabled,
            description=s.description,
            url=s.url,
            lastsync=s.lastsync,
            protocol=s.protocol.version,
            authentication=s.authentication.type,
            transforms=[t.name for t in s.transforms]
        )

    def _init_transform_sets(self):
        for s in self.namelist():
            if s.endswith('.set'):
                self._append_transform_set(s)

    def _append_transform_set(self, s):
        s = TransformSet.parse(self.read(s))
        self.add_transform_set(
            s.name,
            description=s.description,
            transforms=[t.name for t in s.transforms]
        )

    @property
    def icon_files(self):
        return [i for i in self.namelist() if i.startswith('Icons/')]

    @property
    def server_files(self):
        return [self._servers[s]['filename'] for s in self._servers]

    @property
    def entity_files(self):
        return [e for e in self.namelist() if e.endswith('.entity')]

    @property
    def transform_repositories(self):
        return self._match_one(r'TransformRepositories/(.+)?/.*')

    @property
    def transform_sets(self):
        return self._transform_sets.keys()

    @property
    def transform_set_files(self):
        return [self._transform_sets[t]['filename'] for t in self._transform_sets]

    @property
    def entity_categories(self):
        return self._match_one(r'EntityCategories/(.+)?\.category')

    @property
    def machine_files(self):
        return [m for m in self.namelist() if m.endswith('.machine')]

    @property
    def seed_files(self):
        return [s for s in self.namelist() if s.endswith('.seed')]

    @property
    def icon_categories(self):
        return self._match_one(r'Icons/(.+)?/.*', ['Custom'])

    @property
    def transform_files(self):
        return [(t, '%ssettings' % t) for t in self.namelist() if t.endswith('.transform')]

    @property
    def transform_uuids(self):
        return self._match_one(r'TransformRepositories/.*?/(.+)?\.transform')

    def _match_one(self, regex, l=None):
        if l is None:
            l = []
        for r in self.namelist():
            m = re.match(regex, r)
            if not m:
                continue
            category = m.groups()[0]
            if category not in l:
                l.append(category)
        return l

    def get_icons_by_category(self, icon_category_name):
        if icon_category_name == 'Custom':
            return [i for i in self.icon_files if i.count('/') == 1]
        return [i for i in self.icon_files if i.startswith(self.get_icon_category_dir(icon_category_name))]

    def get_transform_repository_dir(self, transform_repository_name):
        return self.path_join(self.transform_repositories_dir, transform_repository_name)

    def get_entity_category_dir(self, entity_category_name):
        name = '/'.join(['EntityCategories', entity_category_name
                        if entity_category_name.endswith('.set') else '%s.set' % entity_category_name])
        if name not in self.transform_set_files:
            self.write_file(name, EntityCategory(name=entity_category_name))
        return name

    def get_transform_set_dir(self, transform_set_name):
        return '/'.join([self.transform_sets_dir,
                         transform_set_name if transform_set_name.endswith('.set') else '%s.set' % transform_set_name])

    def get_icon_category_dir(self, icon_category_name):
        return self.icons_dir if icon_category_name == 'Custom' else self.path_join(self.icons_dir, icon_category_name)

    def add_transform_set(self, transform_set_name, **kwargs):
        transforms = kwargs.get('transforms', [])
        if transform_set_name not in self._transform_sets:
            self._transform_sets[transform_set_name] = dict(
                filename=self.get_transform_set_dir(transform_set_name),
                description=kwargs.get('description', ''),
                transforms=transforms
            )
        elif transforms:
            for t in transforms:
                if t not in self._transform_sets[transform_set_name]['transforms']:
                    self._transform_sets[transform_set_name]['transforms'].append(t)

    def add_transform_repository(self, transform_repository_name):
        if self.mode == 'r':
            raise RuntimeError('Adding a transform repository requires mode "w" or "a"')
        self.get_transform_repository_dir(transform_repository_name)

    def add_entity_category(self, entity_category_name):
        if self.mode == 'r':
            raise RuntimeError('Adding an entity category requires mode "w" or "a"')
        self.get_entity_category_dir(entity_category_name)

    def add_icon_category(self, icon_category_name):
        if self.mode == 'r':
            raise RuntimeError('Adding an icon category requires mode "w" or "a"')
        self.get_icon_category_dir(icon_category_name)

    def write_file(self, zinfo_or_arcname, contents):
        if '\\' in zinfo_or_arcname:
            zinfo_or_arcname.replace('\\', '/')
        if isinstance(contents, Model):
            contents = contents.render(fragment=True, pretty=True)
        self.writestr(zinfo_or_arcname, contents)

    def add_transform_to_set(self, transform_id, transform_set):
        self.add_transform_set(transform_set)
        if transform_id not in self._transform_sets[transform_set]['transforms']:
            self._transform_sets[transform_set]['transforms'].append(transform_id)

    def _write_pending(self):
        # Write pending transform sets
        for transform_set, transform_set_def in self._transform_sets.items():
            transform_set_xml = TransformSet(
                name=transform_set,
                description=transform_set_def['description'],
                transforms=[Transform(name=transform) for transform in transform_set_def['transforms']]
            )
            print('Writing transform set %s to %s...' % (transform_set, self.filename), file=sys.stderr)
            self.write_file(transform_set_def['filename'], transform_set_xml)

        # Write pending server definitions
        for s, server_def in self._servers.items():
            server = MaltegoServer(
                name=s,
                enabled=server_def['enabled'],
                description=server_def['description'],
                url=server_def['url'],
                lastsync=server_def['lastsync'],
                protocol=Protocol(version=server_def['protocol']),
                authentication=Authentication(type_=server_def['authentication']),
                transforms=[Transform(name=transform) for transform in server_def['transforms']]
            )
            print('Writing server %s to %s' % (s, self.filename), file=sys.stderr)
            self.write_file(server_def['filename'], server)

    def add_server(self, server_name, **kwargs):
        transforms = kwargs.get('transforms', [])
        if server_name not in self._servers:
            self._servers[server_name] = dict(
                filename=self.path_join(self.servers_dir, '%s.tas' % server_name),
                enabled=kwargs.get('enabled', True),
                description=kwargs.get('description', 'Local transforms hosted on this machine'),
                url=kwargs.get('url', 'http://localhost'),
                lastsync=kwargs.get('lastsync', time.strftime('%Y-%m-%d')),
                protocol=kwargs.get('protocol', '0.0'),
                authentication=kwargs.get('type_', 'none'),
                transforms=transforms
            )
        elif transforms:
            for t in transforms:
                if t not in self._servers[server_name]['transforms']:
                    self._servers[server_name]['transforms'].append(t)

    def add_transform_to_server(self, server_name, transform_id):
        self.add_server(server_name)
        if transform_id not in self._servers[server_name]['transforms']:
            self._servers[server_name]['transforms'].append(transform_id)

    def remove_transform_from_server(self, server_name, transform_id):
        if transform_id in self._servers[server_name]['transforms']:
            self._servers[server_name]['transforms'].remove(transform_id)

    def add_machine(self, name, contents):
        print('Writing machine %s to %s...' % (name, self.filename), file=sys.stderr)
        self.write_file(self.path_join(self.machines_dir, name), contents)
        self.write_file(self.path_join(self.machines_dir, name.replace('.machine', '.properties')), 'enabled=true')

    def remove_machine(self, filename):
        print('Removing %s from %s...' % (filename, self.machines_dir), file=sys.stderr)
        path = self.path_join(self.machines_dir, os.path.basename(filename))
        self.remove_file(path)
        self.remove_file(path.replace('.machine', '.properties'))

    def remove_file(self, filename):
        self.removefile(filename)

    def merge(self, mtz_file):
        if not isinstance(mtz_file, MtzDistribution):
            raise TypeError('Expected MtzDistribution object got %s instead.' % type(mtz_file).__name__)
        for f in mtz_file.namelist():
            print('Writing %s to %s...' % (f, mtz_file.filename), file=sys.stderr)
            if re.match(r'.*?/(.+)?\.(entity|category|transform(settings)?|mtvs|properties|machine|seed)$', f):
                self.write_file(f, mtz_file.read_file(f))
            elif f.endswith('.tas'):
                self._append_server(mtz_file.read_file(f))
            elif f.endswith('.set'):
                self._append_transform_set(mtz_file.read_file(f))

    def add_entity(self, entity):
        if isinstance(entity, basestring):
            entity = MaltegoEntity.parse(entity)
        if not isinstance(entity, MaltegoEntity):
            raise TypeError('Expected str or MaltegoEntity, not %s', type(entity).__name__)
        self.add_entity_category(entity.category)
        entity_filename = '%s.entity' % self.path_join(self.entities_dir, entity.id)
        print('Installing entity %s to %s...' % (entity.id, entity_filename), file=sys.stderr)
        self.write_file(entity_filename, entity.render(fragment=True, pretty=True))

    def path_join(self, *args):
        return '/'.join(args)

    def read_file(self, filename):
        return self.read(filename)

    def close(self):
        if self.mode != 'r':
            self._write_pending()
        super(MtzDistribution, self).close()


from past.builtins import basestring, long
from future.utils import with_metaclass

from collections import Iterable
from datetime import datetime, date, timedelta
from numbers import Number
import re

from canari.maltego.oxml import MaltegoElement, fields as fields_


__author__ = 'Nadeem Douba'
__copyright__ = 'Copyright 2015, Canari Project'
__credits__ = []

__license__ = 'GPLv3'
__version__ = '0.1'
__maintainer__ = 'Nadeem Douba'
__email__ = 'ndouba@gmail.com'
__status__ = 'Development'

__all__ = [
    'MaltegoException',
    'MaltegoTransformExceptionMessage',
    'MaltegoTransformRequestMessage',
    'Label',
    'MatchingRule',
    'Field',
    'UIMessageType',
    'UIMessage',
    'MaltegoTransformResponseMessage',
    'MaltegoMessage',
    'StringEntityField',
    'EnumEntityField',
    'IntegerEntityField',
    'BooleanEntityField',
    'FloatEntityField',
    'LongEntityField',
    'DateTimeEntityField',
    'DateEntityField',
    'TimeSpan',
    'TimeSpanEntityField',
    'RegexEntityField',
    'ColorEntityField',
    'ArrayEntityField',
    'Entity',
    'Limits'
]


class MaltegoException(MaltegoElement, Exception):
    """MaltegoException is the default container for exception messages."""
    class meta:
        tagname = 'Exception'

    value = fields_.String(tagname='.')
    code = fields_.Integer(attrname='code', required=False)

    def __init__(self, value):
        super(MaltegoElement, self).__init__(value=value)

    def __str__(self):
        return self.value


class MaltegoTransformExceptionMessage(MaltegoElement):
    """
    MaltegoTransformExceptionMessage is the root container for the MaltegoException element.
    """
    exceptions = fields_.List(MaltegoException, tagname='Exceptions')

    def __iadd__(self, exception):
        if isinstance(exception, MaltegoException):
            self.exceptions.append(exception)
        else:
            self.exceptions.append(MaltegoException(str(exception)))
        return self


class Limits(MaltegoElement):
    """
    Limits specifies the soft and hard limits of the MaltegoTransformResponseMessage. The soft and hard limits specify
    the maximum number of entities for response messages.
    """
    soft = fields_.Integer(attrname='SoftLimit', default=500)
    hard = fields_.Integer(attrname='HardLimit', default=10000)


class Label(MaltegoElement):
    """
    Labels are used to convey extra information associated with an Entity in the Maltego user interface. Unlike entity
    fields, labels are only transmitted in response messages and cannot be passed from transform to transform as a
    source of input.
    """
    def __init__(self, name=None, value=None, **kwargs):
        super(Label, self).__init__(name=name, value=value, **kwargs)

    value = fields_.CDATA(tagname='.')
    type = fields_.String(attrname='Type', default='text/text')
    name = fields_.String(attrname='Name')


class MatchingRule(object):
    """Matching rules are used to specify how an entity will be merged in the Maltego user interface. Strict matching
    specifies that an entity will only be merged with another if all it's fields (including the value) are equal.
    Loose matching specifies that two entities will be merged if only the entity values are equal."""
    Strict = "strict"
    Loose = "loose"


class Field(MaltegoElement):
    """Fields are additional sources of input that are attached to an entity."""
    def __init__(self, name=None, value=None, **kwargs):
        super(Field, self).__init__(name=name, value=value, **kwargs)

    name = fields_.String(attrname='Name')
    display_name = fields_.String(attrname='DisplayName', required=False)
    matching_rule = fields_.String(attrname='MatchingRule', default=MatchingRule.Strict, required=False)
    value = fields_.String(tagname='.')


class _Entity(MaltegoElement):
    class meta:
        tagname = 'Entity'

    type = fields_.String(attrname='Type')
    fields = fields_.Dict(Field, key='name', tagname='AdditionalFields', required=False)
    labels = fields_.Dict(Label, key='name', tagname='DisplayInformation', required=False)
    value = fields_.String(tagname='Value')
    weight = fields_.Integer(tagname='Weight', default=1)
    icon_url = fields_.String(tagname='IconURL', required=False)

    def __iadd__(self, other):
        if isinstance(other, Field):
            self.fields[other.name] = other
        elif isinstance(other, Label):
            self.labels[other.name] = other
        return self


class UIMessageType:
    Fatal = "FatalError"
    Partial = "PartialError"
    Inform = "Inform"
    Debug = "Debug"


class UIMessage(MaltegoElement):
    def __init__(self, message=None, **kwargs):
        super(UIMessage, self).__init__(message=message, **kwargs)

    type = fields_.String(attrname='MessageType', default=UIMessageType.Inform)
    message = fields_.String(tagname='.')


class MaltegoTransformResponseMessage(MaltegoElement):
    messages = fields_.List(UIMessage, tagname='UIMessages')
    entities = fields_.List(_Entity, tagname='Entities')

    def __iadd__(self, other):
        if isinstance(other, Entity):
            self.entities.append(other.__entity__)
        elif isinstance(other, _Entity):
            self.entities.append(other)
        elif isinstance(other, UIMessage):
            self.messages.append(other)
        return self


class ValidationError(MaltegoException):
    pass


class StringEntityField(object):

    error_msg = ''

    def __init__(self, name, **extras):
        self.name = name
        self.decorator = extras.pop('decorator', None)
        self.is_value = extras.pop('is_value', False)
        self.display_name = extras.pop('display_name', '')
        self.matching_rule = extras.pop('matching_rule', MatchingRule.Strict)
        self.alias = extras.pop('alias', None)
        self.error_msg = extras.pop('error_msg', self.error_msg)

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        if self.is_value:
            return obj.value
        elif self.name in obj.fields:
            return obj.fields[self.name].value
        elif self.alias in obj.fields:
            return obj.fields[self.alias].value
        return None

    def __set__(self, obj, val):
        if self.is_value:
            obj.value = val
        elif not val:
            if self.name in obj.fields:
                del obj.fields[self.name]
            elif self.alias in obj.fields:
                del obj.fields[self.alias]
        else:
            if self.name not in obj.fields and self.alias not in obj.fields:
                obj.fields[self.name] = Field(
                    name=self.name,
                    value=val,
                    display_name=self.display_name,
                    matching_rule=self.matching_rule
                )
            elif self.name in obj.fields:
                obj.fields[self.name].value = val
            else:
                obj.fields[self.alias].value = val
        if callable(self.decorator):
            self.decorator(obj, val)

    def get_error_msg(self, field, value, **extras):
        return self.error_msg.format(field=field, value=value, **extras)


class EnumEntityField(StringEntityField):

    error_msg = 'Invalid value ({value!r}) set for field {field!r}. Expected one of these values: {expected!r}.'

    def __init__(self, name, choices=None, **extras):
        if not choices:
            raise ValueError('You must specify a non-empty set of choices.')
        self.choices = [str(c) if not isinstance(c, basestring) else c for c in choices]
        super(EnumEntityField, self).__init__(name, **extras)

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        c = super(EnumEntityField, self).__get__(obj, objtype)
        if c is not None and not isinstance(c, basestring):
            c = str(c)
        if c and c not in self.choices:
            raise ValidationError(self.get_error_msg(self.display_name or self.name, c, expected=self.choices))
        return c

    def __set__(self, obj, val):
        val = str(val) if not isinstance(val, basestring) else val
        if val not in self.choices:
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val, expected=self.choices))
        super(EnumEntityField, self).__set__(obj, val)


class IntegerEntityField(StringEntityField):

    error_msg = 'The field value ({value!r}) set for field {field!r} is not an integer.'

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        i = super(IntegerEntityField, self).__get__(obj, objtype)
        try:
            return int(i) if i is not None else None
        except ValueError:
            raise ValidationError(self.get_error_msg(self.display_name or self.name, i))

    def __set__(self, obj, val):
        if not isinstance(val, Number):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val))
        super(IntegerEntityField, self).__set__(obj, val)


class BooleanEntityField(StringEntityField):

    error_msg = 'The field value ({value!r}) set for field {field!r} is not a boolean.'

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        b = super(BooleanEntityField, self).__get__(obj, objtype)
        return b.startswith('t') or b == '1' if b is not None else None

    def __set__(self, obj, val):
        if not isinstance(val, bool):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val))
        super(BooleanEntityField, self).__set__(obj, str(val).lower())


class FloatEntityField(StringEntityField):

    error_msg = 'The field value ({value!r}) set for field {field!r} is not a float.'

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        f = super(FloatEntityField, self).__get__(obj, objtype)
        try:
            return float(f) if f is not None else None
        except ValueError:
            raise ValidationError(self.get_error_msg(self.display_name or self.name, f))

    def __set__(self, obj, val):
        if not isinstance(val, Number):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val))
        super(FloatEntityField, self).__set__(obj, val)


class LongEntityField(StringEntityField):

    error_msg = 'The field value ({value!r}) set for field {field!r} is not a long integer.'

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        l = super(LongEntityField, self).__get__(obj, objtype)
        try:
            return long(l) if l is not None else None
        except ValueError:
            raise ValidationError(self.get_error_msg(self.display_name or self.name, l))

    def __set__(self, obj, val):
        if not isinstance(val, Number):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val))
        super(LongEntityField, self).__set__(obj, val)


class DateTimeEntityField(StringEntityField):

    error_msg = 'The field value ({value!r}) set for field {field!r} is not a valid date time. Date time fields must ' \
                'have the following format: YYYY-MM-DD HH:MM:SS.MS.'

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        d = super(DateTimeEntityField, self).__get__(obj, objtype)
        try:
            return datetime.strptime(d, '%Y-%m-%d %H:%M:%S.%f') if d is not None else None
        except ValueError:
            raise ValidationError(self.get_error_msg(self.display_name or self.name, d))

    def __set__(self, obj, val):
        if not isinstance(val, datetime):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val))
        super(DateTimeEntityField, self).__set__(obj, val.strftime('%Y-%m-%d %H:%M:%S.%f'))


class DateEntityField(StringEntityField):

    error_msg = 'The field value ({value!r}) set for field {field!r} is not a valid date. Date fields must have the ' \
                'following format: YYYY-MM-DD.'

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        d = super(DateEntityField, self).__get__(obj, objtype)
        try:
            return datetime.strptime(d, '%Y-%m-%d').date() if d is not None else None
        except ValueError:
            raise ValidationError(self.get_error_msg(self.display_name or self.name, d))

    def __set__(self, obj, val):
        if not isinstance(val, date):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val))
        super(DateEntityField, self).__set__(obj, val.strftime('%Y-%m-%d'))


class TimeSpan(timedelta):
    matcher = re.compile('(\d+)d (\d+)h(\d+)m(\d+)\.(\d+)s')

    def __str__(self):
        return '%dd %dh%dm%d.%03ds' % (
            abs(self.days),
            int(self.seconds) // 3600,
            int(self.seconds) % 3600 // 60,
            int(self.seconds) % 60,
            int(self.microseconds)
        )

    @classmethod
    def fromstring(cls, ts):
        m = cls.matcher.match(ts)
        if m is None:
            raise ValueError('Time span must be in "%%dd %%Hh%%Mm%%S.%%fs" format')
        days, hours, minutes, seconds, useconds = [int(i) for i in m.groups()]
        return TimeSpan(days, (hours * 3600) + (minutes * 60) + seconds, useconds)


class TimeSpanEntityField(StringEntityField):

    error_msg = 'The field value ({value!r}) set for field {field!r} is not a valid time span. Time spans must have ' \
                'the following format: DDd HHhMMmSS.MSs.'

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        d = super(TimeSpanEntityField, self).__get__(obj, objtype)
        try:
            return TimeSpan.fromstring(d) if d is not None else None
        except ValueError:
            raise ValidationError(self.get_error_msg(self.display_name or self.name, d))

    def __set__(self, obj, val):
        if not isinstance(val, timedelta):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val))
        if val.__class__ is timedelta:
            val = TimeSpan(val.days, val.seconds, val.microseconds)
        super(TimeSpanEntityField, self).__set__(obj, str(val))


class RegexEntityField(StringEntityField):

    error_msg = 'The field value ({value!r}) set for field {field!r} does not match the regular expression /{pattern}/.'

    def __init__(self, name, pattern='.*', **extras):
        super(RegexEntityField, self).__init__(name, **extras)
        self.matcher = re.compile(pattern)

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        v = super(RegexEntityField, self).__get__(obj, objtype)
        if v and not self.matcher.match(v):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, v, pattern=self.matcher.pattern))
        return v

    def __set__(self, obj, val):
        if not isinstance(val, basestring):
            val = str(val)
        if not self.matcher.match(val):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val, pattern=self.matcher.pattern))
        super(RegexEntityField, self).__set__(obj, val)


class ColorEntityField(RegexEntityField):

    error_msg = 'The field value ({value!r}) set for {field!r} is not a valid color (i.e. #000000-#ffffff)'

    def __init__(self, name, **extras):
        super(ColorEntityField, self).__init__(name, pattern='^#[0-9a-fA-F]{6}$', **extras)


class ElementType(object):

    @staticmethod
    def int(value):
        return None if value is None else int(value)

    @staticmethod
    def string(value):
        return None if value is None else str(value).replace('\\,', ',')

    @staticmethod
    def boolean(value):
        return None if value is None else value.startswith('t') or value == '1'

    @staticmethod
    def double(value):
        return None if value is None else float(value)

    @staticmethod
    def float(value):
        return None if value is None else float(value)

    @staticmethod
    def date(value):
        return None if value is None else datetime.strptime(value, '%Y-%m-%d').date()

    @staticmethod
    def matches_type(type_, value):
        return (type_ is ElementType.int and isinstance(value, int)) or \
               (type_ is ElementType.string and isinstance(value, basestring)) or \
               (type_ is ElementType.boolean and isinstance(value, bool)) or \
               ((type_ is ElementType.float or type_ is ElementType.double) and isinstance(value, float)) or \
               (type_ is ElementType.date and isinstance(value, date))


class ArrayEntityField(StringEntityField):

    splitter = re.compile(r'(?<!\\),')

    error_msg = 'The field value ({value!r}) set for {field!r} is not an enumerable type.'
    error_msg_element = 'Could not parse element value ({value!r}) at index {element_pos!r} ' \
                        'as a type of {element_type!r} in field {field!r}.'
    error_msg_array = 'Could not parse input array {value!r} as array of {element_type!r} for field {field!r}.'

    def __init__(self, name, element_type=ElementType.string, **extras):
        super(ArrayEntityField, self).__init__(name, **extras)
        self.element_type = element_type

    def _iter_and_validate(self, val):
        for i, e in enumerate(val):
            if not ElementType.matches_type(self.element_type, e):
                raise ValidationError(self.error_msg_element.format(
                    field=self.display_name or self.name, value=e,
                    element_type=self.element_type.__name__, element_pos=i))
            yield e.replace(',', '\\,') if self.element_type is ElementType.string else str(e)

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        a = super(ArrayEntityField, self).__get__(obj, objtype)
        try:
            return [self.element_type(e) for e in self.splitter.split(a)] if a is not None else None
        except ValueError:
            raise ValidationError(self.error_msg_array.format(
                field=self.display_name or self.name, value=a, element_type=self.element_type.__name__))

    def __set__(self, obj, val):
        if not isinstance(val, Iterable):
            raise ValidationError(self.get_error_msg(self.display_name or self.name, val))
        super(ArrayEntityField, self).__set__(obj, ','.join(self._iter_and_validate(val)))


class EntityTypeFactory(type):
    """The EntityFactory metaclass is responsible for keeping track of and initializing custom entity classes. This is
    required to create entity instances based on the type specifier from the entity XML document."""

    registry = {}

    def __new__(mcs, name, bases, dict_):
        # Just create the class without mangling anything first.
        cls = type.__new__(mcs, name, bases, dict_)

        # We don't need to initialize and register the base Entity class.
        if name == 'Entity':
            return cls

        # This is where we register our custom Entity subclasses.
        # We can register an alias for the entity. This was done to accommodate the difference between V2 and V3
        # entities. This is to support backwards compatibility with older versions of Maltego.
        if '_alias_' not in dict_:
            cls._alias_ = name

        # Create a namespace if it is not explicitly defined. Dynamically created namespaces are based on which module
        # the entity class is defined in.
        if not cls._namespace_:
            # If the entities are coming from the Canari framework then we are going to automatically assign these
            # entities the 'maltego' namespace since they are the Canari definitions of the built-in Maltego entities.
            if cls.__module__ == 'canari.maltego.entities':
                cls._namespace_ = 'maltego'
            # Otherwise, we create the namespace based on the package name
            else:
                cls._namespace_ = cls.__module__.split('.', 1)[0]

        if '_type_' not in dict_:
            cls._type_ = '%s.%s' % (cls._namespace_, name)

        mcs.registry[cls._type_] = cls
        mcs.registry[cls._alias_] = cls

        return cls

    @classmethod
    def create(mcs, entity_type):
        return mcs.registry[entity_type]


class Bookmark:
    NoColor = -1
    Cyan = 0
    Green = 1
    Yellow = 2
    Orange = 3
    Red = 4


class LinkStyle:
    Normal = 0
    Dashed = 1
    Dotted = 2
    DashDot = 3


class LinkLabel:
    UseGlobalSetting = 0
    Show = 1
    Hide = 2


class LinkColor:
    Black = '#000000'
    DarkGray = '#7F7F7F'
    LightGray = '#C3C3C3'
    Red = '#F4291A'
    Orange = '#FF810F'
    DarkGreen = '#30AF44'
    NavyBlue = '#00A2EB'
    Magenta = '##A14DA7'
    Cyan = '#99D9EB'
    Lime = '#B9E500'
    Yellow = '#FFE100'
    Pink = '#FEAFCA'


class Entity(with_metaclass(EntityTypeFactory, object)):

    _namespace_ = None
    _alias_ = None
    _type_ = None
    _category_ = None

    notes = StringEntityField('notes#', display_name='Notes', matching_rule=MatchingRule.Loose)
    bookmark = IntegerEntityField('bookmark#', display_name='Bookmark', choices=range(-1, 5),
                                  matchingrule=MatchingRule.Loose)
    link_label = StringEntityField('link#maltego.link.label', display_name='Link Label',
                                   matching_rule=MatchingRule.Loose)
    link_style = EnumEntityField('link#maltego.link.style', display_name='Link Style',
                                 choices=range(4), matching_rule=MatchingRule.Loose)
    link_thickness = EnumEntityField('link#maltego.link.thickness', display_name='Link Thickness',
                                     choices=range(6), matching_rule=MatchingRule.Loose)
    link_show_label = EnumEntityField('link#maltego.link.show-label', display_name='Link Show Label', choices=range(3),
                                      matching_rule=MatchingRule.Loose)
    link_color = EnumEntityField('link#maltego.link.color', display_name= 'Link Color',
                                 choices=[LinkColor.Black, LinkColor.DarkGray, LinkColor.LightGray, LinkColor.Red,
                                          LinkColor.Orange, LinkColor.DarkGreen, LinkColor.NavyBlue, LinkColor.Magenta,
                                          LinkColor.Cyan, LinkColor.Lime, LinkColor.Yellow, LinkColor.Pink],
                                 matching_rule=MatchingRule.Loose)

    def __init__(self, value='', **kwargs):
        if isinstance(value, _Entity):
            self._entity = value
        else:
            self._entity = _Entity(
                type=kwargs.pop('type', self._type_),
                value=value,
                weight=kwargs.pop('weight', None),
                icon_url=kwargs.pop('icon_url', None),
                fields=self._list_to_dict(kwargs.pop('fields', None)),
                labels=self._list_to_dict(kwargs.pop('labels', None))
            )
        for p in kwargs:
            if hasattr(self, p):
                setattr(self, p, kwargs[p])

    def _list_to_dict(self, obj):
        if not obj:
            return
        if isinstance(obj, dict):
            return obj
        return dict([(i.name, i) for i in obj])

    @property
    def __entity__(self):
        return self._entity

    @property
    def __type__(self):
        return self._type_

    @property
    def fields(self):
        return self._entity.fields

    @property
    def labels(self):
        return self._entity.labels

    @property
    def type(self):
        return self._entity.type

    @type.setter
    def type(self, t):
        self._entity.type = t

    @property
    def value(self):
        return self._entity.value

    @value.setter
    def value(self, v):
        self._entity.value = v

    @property
    def weight(self):
        return self._entity.weight

    @weight.setter
    def weight(self, w):
        self._entity.weight = w

    @property
    def icon_url(self):
        return self._entity.icon_url

    @icon_url.setter
    def icon_url(self, url):
        self._entity.icon_url = url

    def __iadd__(self, other):
        self._entity += other
        return self

    __add__ = __iadd__

    def __getitem__(self, item):
        return self._entity.fields[item].value

    def __setitem__(self, key, value):
        if key not in self._entity.fields:
            self._entity.fields[key] = Field(key, value)
        else:
            self._entity.fields[key].value = value


class MaltegoTransformRequestMessage(MaltegoElement):

    __entities = fields_.List(_Entity, tagname='Entities', required=False)
    _entities = None  # This is so we can cache the transform entity object list.
    _parameters = fields_.Dict(Field, tagname='TransformFields', key='name', required=False)
    limits = fields_.Model(Limits, required=False)

    def __iadd__(self, other):
        if isinstance(other, Entity):
            self.__entities.append(other.__entity__)
        elif isinstance(other, _Entity):
            self.__entities.append(other)
        elif isinstance(other, Field):
            self._parameters[other.name] = other
        elif isinstance(other, Limits):
            self.limits = other
        return self

    @property
    def entity(self):
        """Returns the first Entity object in the transform request.

        :return: first Entity object."""
        if self.entities:
            return self.entities[0]
        return Entity('')

    @property
    def entities(self):
        if not self._entities:
            self._entities = [EntityTypeFactory.create(e.type)(e) for e in self.__entities]
        return self._entities

    @property
    def parameters(self):
        """Returns a list of passed transform parameters in the event that a transform has additional parameters that
        are required in order to operate (i.e. API key). For local transforms, the program arguments are returned
        instead.

        :return: list of parameters."""
        if 'canari.local.arguments' in self._parameters:
            return self._parameters['canari.local.arguments'].value
        return self._parameters


class MaltegoMessage(MaltegoElement):
    """This is the root element for all Maltego request, response, and exception messages that are exchanged between the
    client and the transform server. In the case of local transforms, only response and exception messages are exchanged
    from transform to client."""
    message = fields_.Choice(
        fields_.Model(MaltegoTransformExceptionMessage),
        fields_.Model(MaltegoTransformResponseMessage),
        fields_.Model(MaltegoTransformRequestMessage)
    )



from __future__ import print_function

import sys
from datetime import datetime
from getpass import getuser
from os import path

from mrbob.configurator import Configurator

from canari.commands.common import canari_main
from canari.commands.framework import SubCommand, Argument
import canari


__author__ = 'Nadeem Douba'
__copyright__ = 'Copyright 2015, Canari Project'
__credits__ = []

__license__ = 'GPLv3'
__version__ = '0.1'
__maintainer__ = 'Nadeem Douba'
__email__ = 'ndouba@gmail.com'
__status__ = 'Development'


@SubCommand(
    canari_main,
    help='Creates a Canari transform package skeleton.',
    description='Creates a Canari transform package skeleton.'
)
@Argument(
    'package',
    metavar='<package name>',
    help='The name of the canari package you wish to create.'
)
def create_package(opts):

    package_name = opts.package
    capitalized_package_name = package_name.capitalize()

    variables = {
        'project.name': package_name,
        'entity.example_name': 'My%sEntity' % capitalized_package_name,
        'entity.base_name': '%sEntity' % capitalized_package_name,
        'created.year': datetime.now().year,
        'canari.version': canari.__version__
    }

    defaults = {
        'project.create_example': True,
        'author.name': getuser()
    }

    if not path.exists(package_name):
        print('creating skeleton in %s' % package_name, file=sys.stderr)
        configurator = Configurator('canari.resources.templates:create_package',
                                    package_name,
                                    {'non_interactive': False, 'remember_answers': True},
                                    variables=variables,
                                    defaults=defaults)
        configurator.ask_questions()
        configurator.render()
    else:
        print('A directory with the name %s already exists... exiting' % package_name, file=sys.stderr)
        exit(-1)

    print('done!', file=sys.stderr)
